from collections import defaultdict


class Register:
    def __init__(self):
        self.forward_handles = defaultdict()

    @staticmethod
    def register_forward(model, forward_hooks):
        for hook in forward_hooks:
            model.register_forward_hook(hook)

    @staticmethod
    def register_backward(model, backward_hooks):
        for hook in backward_hooks:
            model.register_backward_hook(hook)





import jittor as jt
import jittor.linalg as linalg
def cal_cov_matrix(x):
    if x.dim() == 2:
        mean_x = jt.mean(x,dim=0)
        x_centered = x - mean_x
        cov_matrix = (x_centered.t() @ x_centered) / (x.shape[0]-1)
        return cov_matrix
    if x.dim() == 3:
        x = x.transpose(0,2).reshape(x.shape[2],-1)
        mean = jt.mean(x,dim=1,keepdims=True)
        centered_input = x - mean
        cov_matrix = centered_input @ centered_input.t() / (x.shape[1] - 1)
        return cov_matrix
    if x.dim() == 4:
        x = x.transpose(0, 2).reshape(x.shape[2], -1)
        mean = jt.mean(x, dim=1, keepdims=True)
        centered_input = x - mean
        cov_matrix = centered_input @ centered_input.t() / (x.shape[1] - 1)
        return cov_matrix

def cal_eig(input):
    eigvals, _ = linalg.eig(input)
    return eigvals




from MMonitor.extensions.extension import Extension


class ForwardInputEigOfCovExtension(Extension):

    def __init__(self):
        super(ForwardInputEigOfCovExtension, self).__init__()
        self._name = 'input_eig_data' 

    def _default(self, module, input, output):
        data = input[0]
        return data

    def _Linear(self, module, input, output):
        data = input[0]
        return data

    def _Conv(self, module, input, output):
        data = input[0]
        b, c, w, h = data.shape
        assert (c > 1), "channel must > 1"
        data = data.transpose(0, 1).contiguous().view(-1, c)
        return data




def linear(interval=1, offset=0):
    docstring = "Track at iterations {" + f"{offset} + n * {interval} " + "| n >= 0}."

    def schedule(global_step):
        shifted = global_step - offset
        return shifted >= 0 and shifted % interval == 0

    schedule.__doc__ = docstring

    return schedule


schedulers = {
    'linear': linear
}


class ScheduleSelector:
   
    @staticmethod
    def select(schedule_name):
        if len(schedule_name) == 0:
            return linear()
        schedule, augs = ScheduleSelector.parse_schedule(schedule_name)
        if schedule not in schedulers:
            raise NotImplementedError(f"hook not found: {schedule}")
        return schedulers[schedule](*augs)
    
    @staticmethod
    def parse_schedule(schedules):
        schedule = schedules[:schedules.index("(")].strip()
        schedule_argument = schedules[schedules.index("(")+1: -1].strip()
        augs = [int(x.strip()) for x in schedule_argument.split(',')]
        return schedule, augs




# 用于计算激活值的norm
from .base_class import SingleStepQuantity
from MMonitor.extensions import ForwardOutputExtension
import jittor as jt

class ForwardOutputGradNorm(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output
        data = jt.flatten(data)
        return jt.norm(data, p=2)  # 使用Jittor的norm计算L2范数

    def backward_extensions(self):
        extensions = [ForwardOutputExtension()]
        return extensions



from .backward_input_extension import BackwardInputExtension
from .backward_output_extension import BackwardOutputExtension
from .backward_output_eig_of_cov_extension import BackwardOutputEigOfCovExtension

__all__ = [
    'BackwardInputExtension',
    'BackwardOutputExtension',
    'BackwardOutputEigOfCovExtension',
]


from .backward_extension import *
from .forward_extension import *
import jittor as jt

class ExtensionSelector:

    @staticmethod
    def select(extension_name):
        if extension_name not in globals():
            raise NotImplementedError(
                "hook not found: {}".format(extension_name))
        extension = globals()[extension_name]
        return extension


from MMonitor.quantity.singlestep.base_class import SingleStepQuantity
from MMonitor.extensions import ForwardInputEigOfCovExtension
from MMonitor.extensions.utils.calculation import cal_cov_matrix,cal_eig
import jittor as jt
from collections import defaultdict

class ForwardInputCovCondition(SingleStepQuantity):
    def _compute(self, global_step):
        eig_values, step = getattr(self._module, 'eig_values', (None, None))
        if eig_values is None or step is None or step != global_step:
            data = self._module.input_eig_data
            cov = cal_cov_matrix(data)
            eig_values = cal_eig(cov)
            if isinstance(eig_values, jt.Var):
                eig_values = eig_values
            else:
                eig_values = jt.Var(eig_values.real) # 直接转换为 Jittor Var
            eig_values = jt.ops.argsort(eig_values, descending=True)  # Sort eigenvalues in descending order
            
            setattr(self._module, 'eig_values', (eig_values, global_step))
        eps = 1e-7
        eig_values_float = eig_values[1]
        condition = eig_values_float[0] / (jt.abs(eig_values_float[-1]) + eps)
        return condition

    def forward_extensions(self):
        extensions = [ForwardInputEigOfCovExtension()]
        return extensions


## ForwardInputExtension
from .forward_input_norm import ForwardInputSndNorm 
from .forward_input_mean import ForwardInputMean
from .forward_input_std import ForwardInputStd
from .forward_input_cov_stable_rank import ForwardInputCovStableRank 
from .forward_input_cov_max_eig import ForwardInputCovMaxEig
from .forward_input_cov_condition import ForwardInputCovCondition
from .forward_input_cov_condition20 import ForwardInputCovCondition20
from .forward_input_cov_condition50 import ForwardInputCovCondition50
from .forward_input_cov_condition80 import ForwardInputCovCondition80
## ForwardOutputExtension 
from .linear_dead_neuron_num import LinearDeadNeuronNum
from .rankme import RankMe
from .zero_activation_precentage import ZeroActivationPrecentage
from .forward_output_grad_norm import ForwardOutputGradNorm
from .forward_output_mean import ForwardOutputMean
from .forward_output_std import ForwardOutputStd
## BackInputExtension
from .back_input_grad_norm import BackInputGradNorm
from .back_input_mean import BackInputMean
from .back_input_std import BackInputStd
## BackOutputExtension ***
from .back_output_grad_norm import BackOutputGradSndNorm
from .back_output_mean import BackOutputMean
from .back_output_std import BackOutputStd
### 
from .weight_norm import WeightNorm
from .attention_save import AttentionSave
from .res_ratio1_save import ResRatio1Save
from .res_ratio2_save import ResRatio2Save


__all__ = [
    'ForwardInputSndNorm',
    'ForwardInputMean',
    'ForwardInputStd',
    'ForwardInputCovStableRank',
    'ForwardInputCovMaxEig',
    'ForwardInputCovCondition',
    'ForwardInputCovCondition20',
    'ForwardInputCovCondition50',
    'ForwardInputCovCondition80',
    'LinearDeadNeuronNum',
    'RankMe',
    'ZeroActivationPrecentage',
    'ForwardOutputGradNorm',
    'ForwardOutputMean',
    'ForwardOutputStd',
    'BackInputGradNorm',
    'BackInputMean',
    'BackInputStd',
    'BackOutputGradSndNorm',
    'BackOutputMean',
    'BackOutputStd',
    'WeightNorm',
    'AttentionSave',
    'ResRatio1Save',
    'ResRatio2Save'
]



from .base_class import SingleStepQuantity
from ...extensions import ForwardOutputExtension
import jittor as jt
# 用于计算死亡神经元的比例，但是设置了一定的大小
class LinearDeadNeuronNum(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.output
        output = data.view(-1, data.shape[-1])
        
        zero_num = [jt.all(output[:, i] <= -2) for i in range(data.shape[-1])]
        return jt.sum(jt.array(zero_num)) / data.shape[-1]  # 计算死神经元比例

    def forward_extensions(self):
        extensions = [ForwardOutputExtension()]
        return extensions


from ..extension import Extension
# 相对于输出的梯度
class BackwardOutputExtension(Extension):

    def __init__(self):
        self._name = 'output_grad'

    def _default(self, module, grad_input, grad_output):
        return grad_output[0]

    def _Linear(self, module, grad_input, grad_output):
        return grad_output[0]

    def _Conv(self, module, grad_input, grad_output):
        return grad_output[0]


from .base_class import SingleStepQuantity
from ...extensions import ForwardInputEigOfCovExtension
import jittor as jt
from ..utils.calculation import *
class ForwardInputCovStableRank(SingleStepQuantity):
    def _compute(self, global_step):
        eig_values, step = getattr(self._module, 'eig_values', (None, None))
        if eig_values is None or step is None or step != global_step:
            data = self._module.input_eig_data
            cov = cal_cov_matrix(data)
            eig_values = cal_eig(cov)
            if isinstance(eig_values, jt.Var):
                eig_values = eig_values
            else:
                eig_values = jt.Var(eig_values.real) # 直接转换为 Jittor Var
            _,eig_values = jt.ops.argsort(eig_values, descending=True)  # 对特征值按降序排序
            setattr(self._module, 'eig_values', (eig_values, global_step))
        if len(eig_values) != 1:
            eig_values = eig_values[1]
            _,eig_values = jt.ops.argsort(eig_values, descending=True)  # 对特征值按降序排序
        max_eigen_value = eig_values[0]
        eigs_sum = eig_values.sum()
        
        if max_eigen_value == 0:
            return eigs_sum * 0
        
        stable_rank = eigs_sum / max_eigen_value
        return stable_rank

    def forward_extensions(self):
        extensions = [ForwardInputEigOfCovExtension()]
        return extensions


import os
import json
import pandas as pd
import jittor as jt
from collections import defaultdict
import numpy as np

class LoadTaskData:
    def __init__(self, root, task) -> None:
        self.task_file_path = os.path.join(root, task)
        self.project_filenames = [f for f in os.listdir(self.task_file_path) if os.path.isdir(os.path.join(self.task_file_path, f))]
        self.project_files_monitor = {}
        self.project_files_train = {}
        self.project_files_val = {}
        self.data = {}
        self._load_project_data()
        
    def _load_project_data(self):
        for project in self.project_filenames:
            dataloader = LoadProjectData(self.task_file_path, project)
            self.project_files_monitor[project] = dataloader.monitor_data
            self.project_files_train[project] = dataloader.train_data
            self.project_files_val[project] = dataloader.val_data
        self.data['monitor'] = self.project_files_monitor
        self.data['train'] = self.project_files_train
        self.data['val'] = self.project_files_val
            
    def load_data(self, quantity_name, data_type='monitor'):
        data = self.data[data_type].copy()
        if quantity_name is not None:
            if isinstance(quantity_name, str):
                data = {quantity_name: data[quantity_name]}
            elif isinstance(quantity_name, list):
                data = {key: data[key] for key in data if key in quantity_name} 
        return self._extract_quantity_data(data,quantity_name = quantity_name)
            
                
    def _get_project_name(self, project_name):
        if isinstance(project_name, str):
            project = self.project_files_monitor.get(project_name, None)
            if project is not None:
                return project_name 
            else:
                return None
        if isinstance(project_name, list):
            result = [self._get_project_name(i) for i in project_name]
            result = [i for i in result if i is not None]
            if len(result) == 0:
                return None
            return result
        
    def _extract_quantity_data(self, data,quantity_name):
        result = {}
        result['legend'] = list(data.keys())
        result_data = defaultdict(dict)
        for i in result['legend']:
            if i == quantity_name:
                df = data[i]
                result_data[i]['title'] = i
                result_data[i]['x_label'] = df.index.name
                result_data[i]['y_label'] = 'value'    
                result_data[i]['x'] = df.index.values
                # 对df[i].values进行改变
                result_len = len(df.index.values)
                values = jt.empty(result_len)
                for j in range(result_len):
                    result_num = df[i].values[j][0]
                    values[j] = result_num
                result_data[i]['y'] = values.numpy()
                result_data[i]['legend'] = i
        return result_data
    
    def get_project_name(self):
        return self.project_filenames
    
    def get_quantity_name(self, project_name=None, data_type='monitor'):
        data = self.data[data_type].copy()
        quantity_data = {}
        quantity_name = set()
        if project_name is not None:
            quantity_data[project_name] = data.get(project_name, None)
            if quantity_data[project_name] is None:
                quantity_data = data
        else:
            quantity_data = data
        for key in quantity_data:
            project_data = quantity_data[key]
            for name in project_data.columns:
                quantity_name.add(name)
        return quantity_name

class LoadProjectData:
    
    def __init__(self, local_path, task) -> None:
        self.local_path = local_path
        self.task = task
        self.monitor_filenames, self.train_filenames, self.val_filenames = self.load_filenames()
        self.monitor_data = self.load_filedata(self.monitor_filenames)
        self.train_data = self.load_filedata(self.train_filenames)
        self.val_data = self.load_filedata(self.val_filenames)
   
    def load_filenames(self):
        local_path = self.local_path
        monitor_filenames = []
        train_filenames = []
        val_filenames = []
        for _, dirnames, _ in os.walk(local_path):  
            for dirname in dirnames:
                current_dir = os.path.join(local_path,dirname)
                for _,_,filenames in os.walk(current_dir):
                    for filename in filenames:  
                        if filename.startswith('monitor'):
                            monitor_filenames.append(os.path.join(current_dir, filename))
                        if filename.startswith('train'):
                            train_filenames.append(os.path.join(current_dir, filename))
                        if filename.startswith('val'):
                            val_filenames.append(os.path.join(current_dir, filename))
        return monitor_filenames, train_filenames, val_filenames


    def load_filedata(self, filenames):
        files = []
        if filenames:
            for i in filenames:
                if self.task in i:
                    with open(i, 'r') as f:
                        files.append(json.load(f))
            df = pd.DataFrame(files) 
            df = df.sort_values(by='step')
            df = df.interpolate()
            df = df.dropna()
            df = df.set_index('step')
            return df
        else:
            return None


    


from MMonitor.quantity.singlestep.base_class import SingleStepQuantity
from MMonitor.extensions import ForwardInputExtension
import jittor as jt

class ForwardInputSndNorm(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.input
        # 由于jittor按照行计算l2范数
        # 需要将其转换为一维
        data_flatten = jt.flatten(data)
        return jt.norm(data_flatten, p=2)  # 使用Jittor的norm计算L2范数

    def forward_extensions(self):
        extensions = [ForwardInputExtension()]
        return extensions




from .base_class import SingleStepQuantity
from ...extensions import ForwardInputExtension

import jittor as jt

class ForwardInputStd(SingleStepQuantity):

    def _compute(self, global_step):
        data = self._module.input
        
        if data.ndim == 3:
            data = data.transpose(0, 2).contiguous().reshape(data.shape[2], -1)
        else:
            data = data.transpose(0, 1).contiguous().reshape(data.shape[1], -1)

        return jt.std(data)  # 使用Jittor的std计算标准差

    def forward_extensions(self):
        extensions = [ForwardInputExtension()]
        return extensions



import json
import yaml


def load_monitor_config(monitor_config):
    # convert yaml or json to dict
    if isinstance(monitor_config, str):
        if monitor_config.endswith('.json'):
            try:
                with open(monitor_config, "r") as file:
                    json_data = json.load(file)
                monitor_config = json_data
            except json.JSONDecodeError as e:
                print(f"JSON 解析错误: {e}")
            except FileNotFoundError:
                print(f"找不到文件: {monitor_config}")
            except Exception as e:
                print(f"发生未知错误: {e}")
        elif monitor_config.endswith('.yaml'):
            try:
                with open(monitor_config, "r") as file:
                    yaml_data = yaml.safe_load(file)
                monitor_config = yaml_data
            except FileNotFoundError:
                print(f"找不到文件: {monitor_config}")
            except yaml.YAMLError as e:
                print(f"YAML 解析错误: {e}")
            except Exception as e:
                print(f"发生未知错误: {e}")
        else:
            print("不支持的文件格式，必须为 .json 或 .yaml")

    if isinstance(monitor_config, dict):
        return monitor_config
    else:
        return None



from .base_class import MultiStepQuantity
from ...extensions import ForwardInputExtension
import jittor as jt

class VarTID(MultiStepQuantity):

    def _compute_ones(self, global_step):
        data = self._module.input
        if data.dim() == 3:
            data = data.transpose(0, 2).contiguous().view(data.shape[2], -1)
        else:
            data = data.transpose(0, 1).contiguous().view(data.shape[1], -1)
        return jt.sqrt(data.var(dim=1))  # 使用Jittor的sqrt和var
    
    def _compute(self, global_step):
        running_sigma = jt.sqrt(self._module.running_var)  # 使用Jittor的sqrt
        diff_data = [d - running_sigma for d in self.cache]
        diff_data = jt.stack(diff_data, dim=0)  # 使用Jittor的stack
        
        eps = 1e-8
        result = diff_data.norm(dim=-1) / (running_sigma.norm(dim=-1) + eps)
        return result.mean()

    def forward_extensions(self):
        extensions = [ForwardInputExtension()]
        return extensions



from collections import defaultdict
import os
import json
import numpy as np
import jittor as jt
import re
class Visualization:
    def __init__(self, monitor=None, dir='./output', project='task', name='name'):
        self.clean_step = 500
        self.dir = dir
        self.monitor = monitor
        if 'class' in str(project):
            self.project = list(project)[0].__name__
        else:
            self.project = type(list(project)[0]).__name__
        self.name = list(name)[0]
        self.save_dir = os.path.join(self.dir, self.project)
        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)

    # Save to local
    def show(self, step, ext=None):
        logs = defaultdict(dict)
        save_logs = defaultdict(dict)
        module_names = self._get_module_name()

        for module_name in module_names:
            quantitis = self.monitor.parse_quantity[module_name]
            quantity_names = self._get_quantity_name(module_name)
            for quantity, quantity_name in zip(quantitis, quantity_names):
                if not quantity.should_show(step):
                    continue
                key = f"{module_name}_{quantity_name}"
                save_dir = os.path.join(self.save_dir, key)
                if not os.path.exists(save_dir):
                    os.makedirs(save_dir)
                val = self._get_result(module_name, quantity_name, step)
                save_logs[key] = val

                if val.size == 1:
                    val = val.item()
                else:
                    val = self._get_result(module_name, quantity_name)

                logs[key] = val

        if ext is not None:
            logs.update(ext)        
        self.save_to_local(step, save_logs)
        return logs
    
    def save_to_local(self, step=0, data_log=None, log_type='monitor'):
        if data_log is not None and len(data_log) != 0:
            self._save(step, data_log, log_type)

    def _save(self, step, data_log, log_type):
        data_log['step'] = step
        data_log = self._apply_save_format(data_log) 
        for task in data_log.keys():
            result_log = defaultdict(dict)
            if task != 'step':
                file_name = os.path.join(self.save_dir, task, f"{log_type}_{step}.json")
                result = data_log[task]
                result_log[task] = result
                result_log['step'] = step
                with open(file_name, 'w') as f:
                    json.dump(result_log, f)

    def log_ext(self, step=None, ext=None, log_type='train'):
        self.save_to_local(step, ext, log_type)

    def _get_module_name(self):
        return self.monitor.get_output().keys()

    def _get_quantity_name(self, module_name):
        return self.monitor.get_output()[module_name].keys()

    def _get_result(self, module_name, quantity_name, step=None):
        if step is not None:
            value = self.monitor.get_output()[module_name][quantity_name][step]
        else:
            value = self.monitor.get_output()[module_name][quantity_name]
        return value
    
    def _apply_save_format(self, value):
        """Apply formatting rules for saved data.

        Jittor tensors are converted to NumPy arrays for saving.

        Args:
            value (Any): Value to be saved.

        Returns:
            Any: Converted value.

        Raises:
            NotImplementedError: If there is no formatting rule for the data type.
        """

        if isinstance(value, jt.Var):

            value = value.numpy()

        elif isinstance(value, dict):

            for key, val in value.items():
                value[key] = self._apply_save_format(val)

        elif isinstance(value, list):

            for idx, val in enumerate(value):
                value[idx] = self._apply_save_format(val)

        elif isinstance(value, np.ndarray):

            value = value.tolist()

        elif isinstance(value, tuple):

            value = tuple(self._apply_save_format(val) for val in value)

        elif isinstance(value, np.float32):

            value = jt.float32(value)

        elif isinstance(value, (float, int)):

            pass

        else:
            raise NotImplementedError(f"No formatting rule for type {type(value)}")

        return value



from .base_class import SingleStepQuantity
from ...extensions import ForwardOutputExtension
from ..utils.calculation import *

import jittor as jt

class RankMe(SingleStepQuantity):
    def _compute(self, global_step):
        data = self._module.output
        data = data.view(data.shape[0], -1)
        svd_values = cal_eig_not_sym(data)  
        eps = 1e-7
        svd_values_norm1 = jt.norm(svd_values, p=1)
        rankme = jt.exp(-jt.sum(
            [(svd_values[i] / svd_values_norm1 + eps) * jt.log(svd_values[i] / svd_values_norm1 + eps) for i in range(min(data.shape))]
        ))
        return rankme

    def forward_extensions(self):
        extensions = [ForwardOutputExtension()]
        return extensions



import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import os

class Figure:
    def __init__(self, data, name, ax=None, default_save_dir='./output/picture'):
        self.x_data = data[name]['x']
        self.y_data = data[name]['y']
        self.x_label = data[name]['x_label']
        self.y_label = data[name]['y_label']
        self.legend = data[name]['legend']
        self.title = data[name]['title']
        self.figsize = (16, 10)
        self.dpi = 80
        self.ax = ax or plt.figure(figsize=self.figsize, dpi=self.dpi).gca()
        self.default_save_dir = default_save_dir
        
    def plot(self, ax=None):
        ax = self._get_ax(ax)
        self._plot(ax)
        self._finalize_plot(ax) 
    
    def show(self, ax=None):
        ax = self._get_ax(ax)
        plt.show()
    
    def save(self, file_name=None, save_dir=None, save_type='png'):
        save_dir = save_dir or self.default_save_dir
        os.makedirs(save_dir, exist_ok=True)
        
        file_name = file_name or f"{self.title}.{save_type}"
        file_path = os.path.join(save_dir, file_name)

        figure = plt.figure(figsize=self.figsize, dpi=self.dpi)
        ax = figure.add_subplot(1, 1, 1)
        self.plot(ax)
        figure.savefig(file_path, format=save_type)
        plt.close(figure)

    def _get_ax(self, ax=None):
        return ax or self.ax
    
    def _plot(self, ax):
        raise NotImplementedError("Subclasses must implement the '_plot' method.")
    
    def _finalize_plot(self, ax=None):
        self._set_xy(ax)
        self._set_title(ax)
        self._remove_borders(ax)
        ax.legend()

    def _set_title(self, ax=None):
        ax.set_title(self.title)
    
    def _set_xy(self, ax=None): 
        ax.set_xlabel(self.x_label)
        ax.set_ylabel(self.y_label)

    def _remove_borders(self, ax=None):
        # Remove borders
        for spine in ax.spines.values():
            spine.set_alpha(0.3)
        ax.spines["top"].set_alpha(0.0)
        ax.spines["right"].set_alpha(0.0)
    
    def unsetlegend(self, ax=None):
        ax = self._get_ax(ax)
        ax.legend().set_visible(False)
